/*
Copyright (C) 2012 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
*/

package org.harctoolbox.IrpMaster;

import com.hifiremote.exchangeir.Analyzer;
import com.hifiremote.exchangeir.RepeatFinder;
import com.hifiremote.exchangeir.UeiLearned;
import com.hifiremote.exchangeir.UeiLearnedImporter;

/**
 * This class interfaces to the ExchangeIR library. It consists entirely of of static functions,
 * using the IrpMaster classes IrSignal, IrSequence as arguments or results.
 *
 */
public class ExchangeIR {
    /** The errlimit parameter in ExchangeIR. Amounts to allowed difference im microseconds. */
    public static final int errLimit = 25;

    /** The number basis for Analyzer outputs. */
    public static final int analyzerBasis = 16;

    /** Number of characters in the hexadecimal representation of UEI learned signals. */
    public static final int ueiLearnedCharsInDigit = 2;

    /** Instantiating this class is meaningless. */
    private ExchangeIR() {}

    /**
     * Returns a new Analyzer, constructed from an IrSignal.
     *
     * @param irSignal input IrSignal
     * @return Analyzer
     */
    public static Analyzer newAnalyzer(IrSignal irSignal) {
        return new Analyzer(irSignal.toIntArray(),
                     irSignal.getIntroBursts(), irSignal.getRepeatBursts(), irSignal.getEndingBursts(),
                     1, (int) irSignal.getFrequency(), errLimit, analyzerBasis);
    }

    /**
     * Returns a new Analyzer, constructed from an IrSequence and a frequency.
     *
     * @param irSequence
     * @param frequency
     * @return Analyzer
     */

    public static Analyzer newAnalyzer(IrSequence irSequence, double frequency) {
        return new Analyzer(irSequence.toInts(), (int) frequency, errLimit, analyzerBasis);
    }

    /**
     * Returns a new Analyzer, constructed from a ModulatedIrSequene.
     * @param irSequence
     * @return Analyzer
     */
    public static Analyzer newAnalyzer(ModulatedIrSequence irSequence) {
        return newAnalyzer(irSequence, irSequence.getFrequency());
    }

    /**
     * Constructs a RepeatFinder from an IrSequence.
     *
     * @param irSequence
     * @return RepeatFinder
     */
    public static RepeatFinder newRepeatFinder(IrSequence irSequence) {
        return new RepeatFinder(irSequence.toInts(), errLimit);
    }

    /**
     * Constructs a UEI learned signal from an IrSignal.
     *
     * @param irSignal
     * @return UEI learned signal
     */
    public static UeiLearned newUeiLearned(IrSignal irSignal) {
        return new UeiLearned(irSignal.toIntArray(), irSignal.getIntroBursts(),
                irSignal.getRepeatBursts(), irSignal.getEndingBursts(), 1, (int) irSignal.getFrequency(), errLimit);
    }

    /**
     * Parses a string containing an UEI learned signal into an IrSignal.
     *
     * @param str
     * @return IrSignal
     */
    public static IrSignal parseUeiLearned(String str) {
        UeiLearnedImporter importer = new UeiLearnedImporter(str);
        return parseUeiLearned(importer);
    }
    
    /**
     * Parses a string containing an UEI learned signal into an IrSignal.
     *
     * @param array Integer array representing signal.
     * @return IrSignal
     */
    public static IrSignal parseUeiLearned(int[] array) {
        UeiLearnedImporter importer = new UeiLearnedImporter(array);
        return parseUeiLearned(importer);
    }

    private static IrSignal parseUeiLearned(UeiLearnedImporter importer) {
        int[] data = importer.getSignal();
        IrSignal irSignal = new IrSignal(importer.getFrequency(), IrpUtils.invalid,
                new IrSequence(data, 0, 2*importer.getNoIntroBursts()),
                new IrSequence(data, 2*importer.getNoIntroBursts(), 2*importer.getNoRepeatBursts()),
                new IrSequence(data, 2*(importer.getNoIntroBursts() + importer.getNoRepeatBursts()), 2*importer.getNoEndingBursts()));
        return irSignal;
    }
 
    /**
     * Tries to interpret the string argument as one of our known formats, and return an IrSignal.
     *
     * @param str String to be interpreted.
     * @return IrSignal, or null on failure.
     */
    public static IrSignal interpretString(String str) {
        int[] array = Pronto.parseString(str);
        try {
            return array != null ? Pronto.ccfSignal(array) : parseUeiLearned(str);
        } catch (IrpMasterException ex) {
            return null;
        }
    }

    /**
     * The main routine is mainly for testing and demonstration.
     *
     * @param args UEI learned signal.
     */
    public static void main(String[] args) {
        StringBuilder str = new StringBuilder();
        for (int i = 0; i < args.length; i++) {
            if (str.length() > 0)
                str.append(" ");
            str.append(args[i]);
        }
        IrSignal irSignal = parseUeiLearned(str.toString());
        System.out.println(irSignal);
        System.out.println(newAnalyzer(irSignal));
        ModulatedIrSequence seq = irSignal.toModulatedIrSequence(7);
        RepeatFinder repeatFinder = newRepeatFinder(seq);
        System.out.println(repeatFinder);
        System.out.println(newAnalyzer(seq));
        DecodeIR.invoke(irSignal);
    }
}
